from NetJin.types import Callable, List, RequestMethod, Dict, Tuple
from NetJin.config import HOST, PORT, DEBUG, BASE_DIR
from NetJin.http.request import Request
from NetJin.http.response import Response

from abc import ABC, abstractmethod
from dataclasses import dataclass
from colorama import init, Fore

import datetime
import threading
import atexit

import socket
import os

__all__ = ["AbstractBase"]
init(True)

_HandleType = Callable[[Request, Response], None]


@dataclass
class _RouteRecordType(object):
    handler: _HandleType
    methods: List[RequestMethod]


@dataclass
class _ErrorRecordType(object):
    handler: _HandleType


class AbstractBase(ABC):
    def __init__(self) -> None:
        self._routes: Dict[str, _RouteRecordType] = {}
        self._error_handlers: Dict[int, _ErrorRecordType] = {}

    def route(
        self, path: str, methods: List[RequestMethod] | None = None
    ) -> Callable[[_HandleType], _HandleType]:
        def wrapper(handler: _HandleType) -> _HandleType:
            route = _RouteRecordType(handler, methods or ["GET"])
            self._routes[path] = route
            return handler

        return wrapper

    def error_route(self, status_code: int) -> Callable[[_HandleType], _HandleType]:
        def wrapper(handler: _HandleType) -> _HandleType:
            route = _ErrorRecordType(handler)
            self._error_handlers[status_code] = route
            return handler

        return wrapper

    def log(self, *messages) -> None:
        date = datetime.datetime.now()
        date = f"{date:[%d of %B, %Y %I:%M:S %p]}"
        if DEBUG:
            print(f"{date:<30}", *messages)

    @abstractmethod
    def handleClient(self, client: socket.socket) -> None:
        raise NotImplementedError("Abstract method must be implemented in child class.")

    def send(
        self,
        client: socket.socket,
        status: Tuple[int, str],
        data: str | bytes,
        content_type: str = "",
    ) -> int:
        if isinstance(data, str):
            response = (
                "HTTP/1.1 %s %s\r\nContent-Type: %s\r\n\r\n" % (*status, content_type)
                + data
            )
            response = response.encode()
        else:
            response = (
                "HTTP/1.1 %s %s\r\nContent-Type: %s\r\n\r\n" % (*status, content_type)
            ).encode() + data
        try:
            status_ = client.send(response)
            client.close()
            return status_
        except OSError as e:
            self.log(f"{Fore.RED}OSError: {Fore.LIGHTRED_EX}{repr(e)}")
            return 0

    def render_error(
        self,
        status: Tuple[int, str],
        response: Response,
        request: Request,
        client: socket.socket,
    ) -> int:
        status_code, status_text = status
        error_template = (
            BASE_DIR
            / "WebServer"
            / "Templates"
            / "Errors"
            / (str(status_code) + ".html")
        )
        temp = self._error_handlers.get(status_code, None)
        if temp:
            temp.handler(request, response)

        if os.path.exists(error_template):
            with open(error_template) as file:
                data = file.read().replace("{{ pathname }}", request.path)
                return self.send(client, status, data, "text/html")
        data = {
            404: "URL '%s' Not Found" % request.path,
            405: "Method [%s] on '%s' Not Allowed" % (request.method, request.path),
        }.get(status_code, "%s - %s" % status)
        return self.send(client, status, data)

    def run(
        self,
        callback: Callable[[socket.socket, str, int], None] | None = None,
        post_callback: Callable[[socket.socket], None] | None = None,
    ) -> None:
        def _close(server: socket.socket) -> None:
            try:
                server.close()
            except OSError:
                ...

        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((HOST, PORT))
        server.listen(5)
        server.settimeout(1)

        post_callback = post_callback if post_callback else _close
        atexit.register(lambda: post_callback(server))

        (
            callback(server, HOST, PORT)
            if callback
            else print("Server listening on http://%s:%s" % (HOST, str(PORT)))
        )

        try:
            while True:
                try:
                    client_socket, _ = server.accept()
                    client_socket.settimeout(60)
                    threading.Thread(
                        target=self.handleClient, args=(client_socket,)
                    ).start()
                except TimeoutError:
                    ...
        except KeyboardInterrupt:
            _close(server)
            exit(0)



from typing import cast
from urllib.parse import urlparse
from dataclasses import dataclass, field
from NetJin.types import TypedDict, Literal, Union, Dict, RequestMethod
import re


__all__ = ["Request", "create_request_object"]

_Headers = TypedDict(
    "_Headers",
    {
        "sec-ch-ua": str,
        "sec-ch-ua-mobile": str,
        "sec-ch-ua-platform": str,
        "Sec-Fetch-User": str | None,
        "Sec-Fetch-Site": str,
        "Sec-Fetch-Mode": str,
        "Sec-Fetch-Dest": str,
        "Accept": str,
        "User-Agent": str,
        "Referer": str | None,
        "Accept-Encoding": str,
        "Accept-Language": str,
        "Cache-Control": str | None,
        "Upgrade-Insecure-Requests": str | None,
        "Cookie": Dict[str, str],
    },
)

ConnectionType = Literal["keep-alive", "close"]


@dataclass
class Request:
    url: str

    path: str = field(init=False)
    scheme: str = field(init=False)
    netloc: str = field(init=False)
    params: Dict[str, str] = field(init=False)
    queries: Dict[str, str] = field(init=False)
    fragment: Dict[str, str] = field(init=False)

    Host: str
    Connection: ConnectionType
    headers: Union[_Headers, Dict[str, str]]
    method: RequestMethod

    user_parameters: Dict[str, str] | None = field(init=False, default=None)

    def __post_init__(self):
        parser = urlparse(self.url)
        self.path = parser.path
        self.scheme = parser.scheme
        self.netloc = parser.netloc
        self.params = dict(re.findall(r";([^=;]+)=([^=;]+)", parser.params))
        self.queries = dict(re.findall(r"([^&=]+)=([^&=]*)", parser.query))
        self.fragment = dict(re.findall(r"#([^#]+)", parser.fragment))


def create_request_object(network_content: str) -> Request:
    network_content_obj: Dict[str, str] = dict(
        re.findall(r"([\w-]+): ([^\n]+)", network_content)
    )
    host = network_content_obj.get("Host", "").rstrip("\r")
    connection = network_content_obj.get("Connection", "").rstrip("\r")
    cookie = dict(
        re.findall(r"([^=;]+)=([^=;]+)", network_content_obj.get("Cookie", ""))
    )

    headers = network_content.split("\r\n")
    request_line = headers[0]
    request_query = request_line.split()

    method, path, _ = request_query if len(request_query) == 3 else (None, "", None)
    try:
        del network_content_obj["Host"]
        del network_content_obj["Connection"]
    except KeyError:
        ...
    return Request(
        Host=host,
        Connection=cast(ConnectionType, connection),
        headers=cast(
            _Headers | Dict[str, str],
            {
                **network_content_obj,
                "Cookie": cookie,
                "Cache_Control": network_content_obj.get("Cache-Control", None),
                "Upgrade_Insecure_Requests": network_content_obj.get(
                    "Upgrade-Insecure-Requests", None
                ),
                "Referer": network_content_obj.get("Referer", None),
                "Sec_Fetch_User": network_content_obj.get("Sec-Fetch-User", None),
            },
        ),
        method=cast(RequestMethod, method.upper() if method else "GET"),
        url="http://" + host + path,
    )



from http import HTTPStatus
from jinja2 import (
    Environment,
    FileSystemLoader,
    ChoiceLoader,
    Template,
    TemplateNotFound,
)
from colorama import Fore, init
from typing import Any, Dict, List
import socket
import json

TEMPLATE_DIRS = None
try:
    exec("from settings import TEMPLATE_DIRS")
except ModuleNotFoundError:
    from NetJin.config.settings import TEMPLATE_DIRS

init(True)

__all__ = ["Response"]

if TEMPLATE_DIRS is None or isinstance(TEMPLATE_DIRS, str):
    template_loader = FileSystemLoader(TEMPLATE_DIRS if TEMPLATE_DIRS else "templates")
else:
    template_loader = ChoiceLoader(
        [FileSystemLoader(template) for template in TEMPLATE_DIRS]
    )
env = Environment(loader=template_loader)


class Response:
    def __init__(self, client: socket.socket) -> None:
        self.__client = client

    def render(self, template: str | Template | List[str | Template], **context: Any) -> int:
        try:
            _template = env.get_or_select_template(template)
            rendered_html = _template.render(context)
            return self.send(
                rendered_html, 200, {"Content-Type": "text/html; charset=utf-8"}
            )
        except TemplateNotFound:
            return self.send("File Not Found", 404, headers={"Connection": "close"})

    def end(self, status: int) -> int:
        return self.send("", status, headers={"Connection": "close"})

    def send(
        self,
        content: Any,
        status: int | None = None,
        headers: Dict[str, str] | None = None,
    ) -> int:
        headers = headers or {}
        status = status or 200

        data: str = content
        if isinstance(content, str):
            self.add_header(headers, "Content-Type", "text/plain; charset=utf-8")
        else:
            try:
                data = json.dumps(content)
                if isinstance(content, (dict, list)):
                    self.add_header(
                        headers, "Content-Type", "application/json; charset=utf-8"
                    )
            except (TypeError, ValueError):
                self.add_header(headers, "Content-Type", "text/plain; charset=utf-8")
                data = "Cannot parse data"
        try:
            status_phrase = HTTPStatus(status).phrase
        except ValueError:
            status_phrase = "Unknown Status Code"

        response = self.__prepare_response(
            data.encode(), f"{status} {status_phrase}", headers
        )
        try:
            # status_ = self.__client.sendall(response)
            status_ = self.__client.send(response)
            # self.__client.close()
            return status_
        except OSError as e:
            print(f"{Fore.RED}OSError: {Fore.LIGHTRED_EX}{repr(e)}")
            return 0

    def add_header(
        self, headers: Dict[str, str], header: str, value: str, force: bool = False
    ) -> Dict[str, str]:
        keys = [key.lower() for key in headers.keys()]
        if force:
            headers[header] = value
        elif header.lower() not in keys:
            headers[header] = value
        return headers

    def __prepare_response(
        self,
        content: bytes,
        status: str = "200 OK",
        headers: Dict[str, str] | None = None,
    ) -> bytes:

        headers = headers or {
            "X-Content-Type-Options": "nosniff",
            "Connection": "keep-alive",
        }

        header_response = "\r\n".join(f"{key}: {value}" for key, value in headers.items())
        content_length = len(content)

        response = (
            f"HTTP/1.1 {status}\r\n"
            f"Content-Length: {content_length}\r\n"
            f"{header_response}\r\n\r\n"
        )
        print(response.rstrip(), "\n")
        response = response.encode()

        if isinstance(content, str):
            content = content.encode()
        return response + content

    def set(self, key: str, value: Any) -> None:
        setattr(self, key, value)

    def get(self, key: str, default: Any) -> Any:
        return getattr(self, key, default)



import importlib.util
import os
import pathlib
from .settings import *  # noqa: F403


def load_config():
    conf_path = pathlib.Path(os.getcwd()) / "config" / "settings.py"
    module_name = "settings"
    if conf_path.exists():
        spec = importlib.util.spec_from_file_location(module_name, conf_path)
        if spec:
            user_conf = importlib.util.module_from_spec(spec)
            if spec.loader:
                spec.loader.exec_module(user_conf)

            for key in dir(user_conf):
                if not key.startswith("__"):
                    globals()[key] = getattr(user_conf, key)


load_config()


from typing import Literal, List, Tuple, Callable, Any
import threading
import socket
import re
import os


def render(filename: str) -> str:
    path = os.path.join("template", filename)
    with open(path) as file:
        return file.read()


class WebSocket:
    def __init__(self, host: str, port: int) -> None:
        self.host = host
        self.port = port
        self.routes = {}

    def route(
        self, path: str, methods: List[Literal["GET", "POST", "PUT", "PATCH", "DELETE"]]
    ):
        def wrapper(func: Callable[[Any], Tuple[str, int]]):
            self.routes[path] = {"handler": func, "methods": methods}
            return func

        return wrapper

    def handle_client(self, client_socket: socket.socket) -> None:
        request_data = client_socket.recv(1024).decode()
        headers = request_data.split("\r\n")
        request_line = headers[0]
        method, path, _ = request_line.split()
        print(method, path)
        if method != "GET":
            client_socket.close()
            return
        for route, route_info in self.routes.items():
            if re.match(f"^{route}$", path):
                response_body, status_code = route_info["handler"]()
                response = f"HTTP/1.1 {status_code}\r\n\r\n{response_body}".encode()
                client_socket.send(response)
                client_socket.close()
                return
        else:
            filename = os.path.join("static/", path.lstrip("/"))
            if os.path.exists(filename) and os.path.isfile(filename):
                with open(filename, "rb") as file:
                    response = b"HTTP/1.1 200 OK\r\n\r\n" + file.read()
                    client_socket.send(response)
                    client_socket.close()
            else:
                response = b"HTTP/1.1 404 Not Found\r\n\r\nFile Not Found"
                client_socket.send(response)
                client_socket.close()

        client_socket.close()

    def run(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind((self.host, self.port))
        server_socket.listen(5)
        print(f"Server listening on http://{self.host}:{self.port}")

        while True:
            try:
                client_socket, _ = server_socket.accept()
                client_handler = threading.Thread(
                    target=self.handle_client, args=(client_socket,)
                )
                client_handler.start()
            except KeyboardInterrupt:
                server_socket.close()
                break


app = WebSocket("127.0.0.1", 5500)


@app.route(path="/", methods=["GET"]) # type: ignore
def index():
    return "<h1>Hello World</h1>", 200


@app.route(path="/home", methods=["GET"]) # type: ignore
def home():
    return render("index.html"), 200


@app.route(path="/name", methods=["GET"]) # type: ignore
def return_json():
    import json

    obj = {"name": "Sachin Acharya", "age": 24}

    data = json.dumps(obj)
    return data, 200


if __name__ == "__main__":
    app.run()



from NetJin.types import TypedDict, Literal, Union, Dict, RequestMethod
from dataclasses import dataclass, field

_Headers = TypedDict(
    "_Headers",
    {
        "sec-ch-ua": str,
        "sec-ch-ua-mobile": str,
        "sec-ch-ua-platform": str,
        "Sec-Fetch-User": str | None,
        "Sec-Fetch-Site": str,
        "Sec-Fetch-Mode": str,
        "Sec-Fetch-Dest": str,
        "Accept": str,
        "User-Agent": str,
        "Referer": str | None,
        "Accept-Encoding": str,
        "Accept-Language": str,
        "Cache-Control": str | None,
        "Upgrade-Insecure-Requests": str | None,
        "Cookie": Dict[str, str],
    },
)

ConnectionType = Literal["keep-alive", "close"]

@dataclass
class Request:
    """This class is wrapper for the incomming Http Request from the client.
    This also extract all the data received from the client for easier access.
    """

    url: str

    path: str = field(init=False)
    scheme: str = field(init=False)
    netloc: str = field(init=False)
    params: Dict[str, str] = field(init=False)
    queries: Dict[str, str] = field(init=False)
    fragment: Dict[str, str] = field(init=False)

    Host: str
    Connection: ConnectionType
    headers: Union[_Headers, Dict[str, str]]
    method: RequestMethod

    user_parameters: Dict[str, str] | None = field(init=False, default=None)



from .Types import *
from .Settings import *
from functools import wraps
import socket
import datetime
import re
import os
import threading
import json
import keyboard
import sys
from dataclasses import dataclass


__all__ = ["render_template", "WebSocket", "jsonify"]


def jsonify(data: dict):
    return json.dumps(data)


def render_template(filename) -> str:
    with open(os.path.join(TEMPLATE_FOLDER, filename)) as file:
        return file.read()


@dataclass
class ContentType:
    HTML: str = "text/html"
    CSS: str = "text/css"
    JAVASCRIPT: str = "application/javascript"
    JSON: str = "application/json"
    OCTSTREAM: str = "application/octet-stream"  # ? For Binary

    def get(
        self,
        key: Literal["HTML", "CSS", "JAVASCRIPT", "JSON", "OCTSTREAM"],
        default: str = "application/octet-stream",
    ) -> Any:
        return getattr(self, key, default)


class WebSocket:
    def __init__(
        self, host: str = "127.0.0.1", port: int = 5500, debug: bool = False
    ) -> None:
        self._host = host
        self._port = port
        self._routes = {}
        self._error_handles = {}
        self._isDebug = debug
        self._ContentType = ContentType()

    def route(self, path: str, methods: List[RequestType] | None = None):
        def wrapper(func: Callable):
            self._routes[path] = {
                "handler": func,
                "methods": methods if methods else ["GET"],
            }
            return func

        return wrapper

    def handleError(self, status_code: int):
        def wrapper(func: Callable):
            self._error_handles[status_code] = {
                "handler": func,
            }

    def __log(self, *message):
        date = datetime.datetime.now()
        date = f"{date:[%d of %B, %Y %I:%M:%S %p]}"
        if self._isDebug:
            print(f"{date:<30}", *message)

    def __handleClient(self, client_socket: socket.socket) -> None:
        user_requests = client_socket.recv(1024).decode()
        if self._isDebug:
            print(user_requests)
            status = re.findall(r"([\w-]+): ([^\n]+)", user_requests)
            # print(dict(status))
        headers = user_requests.split("\r\n")
        request_line = headers[0]
        request_query = request_line.split()
        if not len(request_query) == 3:
            client_socket.close()
            return
        method, path, _ = request_query
        self.__log(method, path)

        if method.lower() != "get":
            client_socket.close()
            return

        # Handle Routing
        for route, route_info in self._routes.items():
            if re.match(f"^{route}$", path):
                if not "GET" in route_info["methods"]:
                    self.__send(
                        403,
                        "Not Allowed",
                        "Method '%s' on route '%s' is not allowed" % (method, path),
                        client_socket,
                    )
                    return

                handler_res = route_info["handler"]()
                status_code = 200
                response_body = handler_res
                if isinstance(handler_res, list) or isinstance(handler_res, tuple):
                    response_body, status_code = handler_res
                response = f"HTTP/1.1 {status_code}\r\nContent-Type: {self._ContentType.HTML}\r\n\r\n{response_body}".encode()
                client_socket.send(response)
                client_socket.close()
                return

        # Handling Static Content
        for static in STATIC_FOLDERS:
            filename = os.path.join(static, path.lstrip("/"))
            if os.path.exists(filename) and os.path.isfile(filename):
                with open(filename, "rb") as file:
                    attr = os.path.splitext(filename)[1].lstrip(".")
                    response = b"HTTP/1.1 200 OK\r\n" + self._ContentType.get(attr.upper()).encode() + b"\r\n\r\n" + file.read()
                    client_socket.send(response)
                    client_socket.close()
                    return

        # Return 404
        if 404 in self._error_handles:
            response = self._error_handles["handler"]
            response = b"HTTP/1.1 404 Not Found\r\n\r\n" + response.encode()
            client_socket.send(response)
            client_socket.close()
            return
        response = b"HTTP/1.1 404 Not Found\r\n\r\nFile Not Found"
        client_socket.send(response)
        client_socket.close()

    def __send(
        self, status_code: int, status_name: str, data: str, client: socket.socket
    ):
        # No ContentType added - change later
        response = "HTTP/1.1 %s %s\r\n\r\n%s" % str(status_code), status_name, data
        client.send(response)
        client.close()

    def run(self) -> NoReturn:
        import atexit

        def close(server):
            try:
                server.close()
            except OSError:
                ...
            sys.exit(0)

        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((self._host, self._port))
        server.listen(5)

        atexit.register(lambda: close(server))
        print("Server listening on http://%s:%s" % (self._host, self._port))

        while True:
            try:
                try:
                    client_socket, _ = server.accept()
                except SystemExit:
                    ...

                # self.__log("Connected to tcp://%s:%s" % _)

                threading.Thread(
                    target=self.__handleClient, args=(client_socket,)
                ).start()
            except KeyboardInterrupt:
                server.close()
                break



import socket
from typing import Any, Dict, List
from jinja2 import Template

__all__ = ["Response"]


class Response:
    """Http Response Object. This is the class that can be used to send response back to the client."""
    def __init__(self, client: socket.socket) -> None: ...

    def render(self, template: str | Template | List[str | Template], **context: Any) -> int:
        """Render HTML (Jinja) template file. This method pre-renders the
        HTML file and serve it making loading of webpages faster.

        Args:
            template (str): Name of the template to be rendered. Directory
                is obtained from TEMPLATE_DIRS const of setings.py file. This settngs.py
                file in the workspace is prioritized over the modules settings.py file.
        """
        ...

    def end(self, status: int) -> int:
        """Respond with blank message

        Args:
            status (int): Http Status for the response.

        Returns:
            int: Amount of bytes sent.
        """
        ...

    def send(
        self,
        data: Any,
        status: int | None = None,
        headers: Dict[str, str] | None = None,
    ) -> int:
        """Send response back to the client.

        Args:
            content (Any): Data or Message that is to be responded back to client.
            status (int): Http status code for response.
            headers (Dict[str, str] | None, optional): Http Response Headers. Defaults to None.

        Returns:
            int: Amount of bytes transfered to the client.
        """
        ...

    def add_header(
        self, headers: Dict[str, str], header: str, value: str, force: bool = False
    ) -> Dict[str, str]:
        """Add header with value to the headers dictionary. This is simple logic that adds header
        if given key is not already present in the dictionary.

        Args:
            headers (Dict[str, str]): Http Headers that will be sent to the client.
            header (str): Http Header that is to be added to given dictionary of headers.
            value (str): Value for the 'header'.
            force (bool, optional): Make sure the value of header is updated if the key already exists. Defaults to False.
        """
        ...

    def set(self, key: str, value: Any) -> None:
        "Set attribute to the response class"
        ...

    def get(self, key: str, default: Any) -> Any:
        "Get attribute from the response class"
        ...



from typing import (
    Literal,
    Tuple,
    Callable,
    NoReturn,
    Self,
    TypedDict,
    TypeVar,
    Union,
    List,
    Any,
    Dict,
)

__all__ = [
    "Literal",
    "Tuple",
    "Callable",
    "NoReturn",
    "Self",
    "TypedDict",
    "TypeVar",
    "Union",
    "List",
    "Any",
    "Dict",
    "RequestMethod",
]

RequestMethod = Literal["GET", "POST", "PUT", "DELETE", "PATCH", "UPDATE"]



from NetJin.utils import extract_route_pattern
from NetJin.http.response import Response
from NetJin.http.request import Request, create_request_object  # type: ignore
from NetJin.types import Callable, List, RequestMethod, Dict, Tuple
from NetJin.config import STATIC_DIRS, BASE_DIR, HOST, PORT, DEBUG

import socket
from dataclasses import dataclass
from colorama import init, Fore
import datetime
import threading
import atexit
import mimetypes
import os

__all__ = ["WebServer"]
init(True)

_HandleType = Callable[[Request, Response], None]


@dataclass
class _RouteRecordType(object):
    handler: _HandleType
    methods: List[RequestMethod]


@dataclass
class _ErrorRecordType(object):
    handler: _HandleType

class WebServer(object):
    def __init__(self, debug: bool = True) -> None:
        self._routes: Dict[str, _RouteRecordType] = {}
        self._error_handlers: Dict[int, _ErrorRecordType] = {}
        self._isDebug = debug
    
    def route(
        self, path: str, methods: List[RequestMethod] | None = None
    ) -> Callable[[_HandleType], _HandleType]:
        def wrapper(handler: _HandleType) -> _HandleType:
            route = _RouteRecordType(handler, methods or ["GET"])
            self._routes[path] = route
            return handler

        return wrapper

    def error_route(self, status_code: int) -> Callable[[_HandleType], _HandleType]:
        def wrapper(handler: _HandleType) -> _HandleType:
            route = _ErrorRecordType(handler)
            self._error_handlers[status_code] = route
            return handler

        return wrapper

    def log(self, *messages) -> None:
        date = datetime.datetime.now()
        date = f"{date:[%d of %B, %Y %I:%M:S %p]}"
        if DEBUG:
            print(f"{date:<30}", *messages)
    
    def handleClient(self, client: socket.socket) -> None:
        user_requests = client.recv(1024).decode()
        request = create_request_object(user_requests)

        if not request.method or not request.path:
            client.close()
            return

        if self._isDebug:
            self.log(request.method, request.path)

        # Handling Routing
        for route, route_info in self._routes.items():
            route_ = extract_route_pattern(route, request.path)
            if isinstance(route_, dict):
                response = Response(client)
                if request.method not in route_info.methods:
                    temp = self._error_handlers.get(405, None)
                    if temp:
                        temp.handler(request, response)
                        return

                    self.send(
                        client,
                        (405, "Not Allowed"),
                        "Method '%s' on route '%s' not allowed"
                        % (request.method, request.path),
                    )
                    return

                if not request.user_parameters:
                    request.user_parameters = route_
                else:
                    request.user_parameters.update(route_)

                route_info.handler(request, response)
                return

        # Handling Static Content
        for static in STATIC_DIRS:
            filename = os.path.join(static, request.path.lstrip("/"))
            if os.path.exists(filename) and os.path.isfile(filename):
                with open(filename, "rb") as file:
                    mimetype, _ = mimetypes.guess_type(filename)
                    if mimetype is None:
                        mimetype = ""
                    #! Change static serve; add cache control for static
                    self.send(client, (200, "OK"), file.read(), mimetype)
                    return

        #! Handle Error Condition (Make user configurable)
        temp = self._error_handlers.get(404, None)
        if temp:
            response = Response(client)
            temp.handler(request, response)
            return

        filepath = BASE_DIR / "views" / "errors" / "NotFound.html"
        with open(filepath) as file:
            self.send(
                client,
                (404, "Not Found"),
                file.read().replace("{{ pathname }}", request.path),
                "text/html",
            )
    def send(
        self,
        client: socket.socket,
        status: Tuple[int, str],
        data: str | bytes,
        content_type: str = "",
    ) -> int:
        if isinstance(data, str):
            response = (
                "HTTP/1.1 %s %s\r\nContent-Type: %s\r\n\r\n" % (*status, content_type)
                + data
            )
            response = response.encode()
        else:
            response = (
                "HTTP/1.1 %s %s\r\nContent-Type: %s\r\n\r\n" % (*status, content_type)
            ).encode() + data
        try:
            status_ = client.send(response)
            client.close()
            return status_
        except OSError as e:
            self.log(f"{Fore.RED}OSError: {Fore.LIGHTRED_EX}{repr(e)}")
            return 0

    def render_error(
        self,
        status: Tuple[int, str],
        response: Response,
        request: Request,
        client: socket.socket,
    ) -> int:
        status_code, status_text = status
        error_template = (
            BASE_DIR
            / "views"
            / "errors"
            / (str(status_code) + ".html")
        )
        temp = self._error_handlers.get(status_code, None)
        if temp:
            temp.handler(request, response)

        if os.path.exists(error_template):
            with open(error_template) as file:
                data = file.read().replace("{{ pathname }}", request.path)
                return self.send(client, status, data, "text/html")
        data = {
            404: "URL '%s' Not Found" % request.path,
            405: "Method [%s] on '%s' Not Allowed" % (request.method, request.path),
        }.get(status_code, "%s - %s" % status)
        return self.send(client, status, data)

    def run(
        self,
        callback: Callable[[socket.socket, str, int], None] | None = None,
        post_callback: Callable[[socket.socket], None] | None = None,
    ) -> None:
        def _close(server: socket.socket) -> None:
            try:
                server.close()
            except OSError:
                ...

        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((HOST, PORT))
        server.listen(5)
        server.settimeout(1)

        post_callback = post_callback if post_callback else _close
        atexit.register(lambda: post_callback(server))

        (
            callback(server, HOST, PORT)
            if callback
            else print("Server listening on http://%s:%s" % (HOST, str(PORT)))
        )

        try:
            while True:
                try:
                    client_socket, _ = server.accept()
                    client_socket.settimeout(60)
                    threading.Thread(
                        target=self.handleClient, args=(client_socket,)
                    ).start()
                except TimeoutError:
                    ...
        except KeyboardInterrupt:
            _close(server)
            exit(0)


from pathlib import Path
from typing import List

__all__ = [
    "BASE_DIR",
    "HOST",
    "PORT",
    "DEBUG",
    "ALLOWED_HOST",
    "TEMPLATE_DIRS",
    "STATIC_DIRS",
    "PUBLIC_DIR",
    "MEDIA_DIR",
]

BASE_DIR = Path(__file__).resolve().parent.parent

# Server Configurations
HOST: str = "127.0.0.1"
PORT: int = 5500
DEBUG: bool = True
ALLOWED_HOST: List[str] = []

# Views and Static
TEMPLATE_DIRS: str | List[str] = "views"
STATIC_DIRS: List[str] = ["static"]
PUBLIC_DIR: str = "public"
MEDIA_DIR: str = "media"



from NetJin.http.response import Response
from NetJin.http.request import Request
from NetJin.types import Callable, List, RequestMethod, Tuple

import socket
from dataclasses import dataclass

__all__ = ["WebServer"]

_HandleType = Callable[[Request, Response], None]

@dataclass
class _RouteRecordType(object):
    handler: _HandleType
    methods: List[RequestMethod]

@dataclass
class _ErrorRecordType(object):
    handler: _HandleType

class WebServer(object):
    def __init__(self, debug: bool = True) -> None:
        """Create an instance of a WebServer

        Args:
            debug (bool, optional): Use this for development. Defaults to True.
        """

    def route(
        self, path: str, methods: List[RequestMethod] | None = None
    ) -> Callable[[_HandleType], _HandleType]:
        """Create a Route using this decorator.

        Example:
            @route("/", methods=['GET', 'POST'])

            def home(Response response, Request request):
                return request.send("Hello, World!")

        Args:
            path (str): URL Fragment to which, the handler function is executed.
            methods (List[RequestMethod] | None, optional): Allowed HTTP Methods to which this method is triggered. Defaults to None.

        Returns:
            Callable[[_HandleType], _HandleType]: Callable method that handles the routing specific operations.
        """
        ...

    def error_route(self, status_code: int) -> Callable[[_HandleType], _HandleType]:
        """Handle HTTP Error routing

        Args:
            status_code (int): StatusCode for which this route-handler will be executed.

        Returns:
            Callable[[_HandleType], _HandleType]: Callable handler function.
        """
        ...

    def log(self, *messages) -> None:
        """Logging out messages to the console."""
        ...

    def send(
        self,
        client: socket.socket,
        status: Tuple[int, str],
        data: str | bytes,
        content_type: str = "",
    ) -> int:
        """Sends response back to the client.

        Args:
            client (socket.socket): Client socket to which response is to be sent.
            status (Tuple[int, str]): HTTP Status Code to send.
            data (str): Content of data to send.
            content_type (str, optional): ContentType header for response. Defaults to "".

        Returns:
            int: Amount of bytes sent.
        """
        ...

    def render_error(
        self,
        status: Tuple[int, str],
        response: Response,
        request: Request,
        client: socket.socket,
    ) -> int:
        """Handles error_route() pipelining.

        Args:
            status (Tuple[int, str]): StatusCode for which pipelining is to be added.
            response (Response): Response object to send to handler function.
            request (Request): Request object to send to handler function.
            client (socket.socket): Client socket to which response is to be sent.

        Returns:
            int: Amount of bytes sent.
        """
        ...

    def run(
        self,
        callback: Callable[[socket.socket, str, int], None] | None = None,
        post_callback: Callable[[socket.socket], None] | None = None,
    ) -> None:
        """Start the server and subsequently handles incomming requests.

        Args:
            callback (((server: socket, host: str, port: int) -> None) | None, optional): Callback Function. It will be called when server starts listening. Defaults to None.
            post_callback (((server: socket) -> None) | None, optional): Callback function which will be called when server is turn off (intentionally or unintentionally) to release some resources. Defaults to None.
        """
        ...



import re
from typing import Any, Dict
from urllib.parse import unquote


__all__ = ["extract_route_pattern"]


def extract_route_pattern(pattern: str, string: str) -> Dict[str, str | Any] | None:
    # Removing browser quotes like %20 representing space.
    string = unquote(string)
    # Escape special characters in pattern
    pattern = re.escape(pattern)
    # Add optional trailing slash to the pattern
    pattern += r"/*$"
    # Replace <variable_name> with named capturing group regex
    pattern = re.sub(r"<(\w+)>", r"(?P<\1>[^/]+)", pattern)
    # Compile the regex pattern
    pattern_regex = re.compile(pattern)
    # Match the pattern against the string
    match = pattern_regex.match(string)
    # Extract variable values
    if match:
        return match.groupdict()
    else:
        return None


if __name__ == "__main__":
    # Test cases
    pattern = "/path/<name>"
    string = "/path/openai"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? Output: {'name': 'openai'}

    pattern = "/<name>:<age>"
    string = "/openai:12"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? Output: {'name': 'openai', 'age': '12'}

    pattern = "/path/next/<name>/<age>"
    string = "/path/next/openai/24"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? Output: {'name': 'openai', 'age': '24'}

    pattern = "/path/sachin/s/a/c/h/i/n/<power>/name/<age>"
    string = "/path/sachin/s/a/c/h/i/n/500/name/24"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? Output: {'power': '500', 'age': '24'}

    pattern = "<username>@<domain>.com"
    string = "sachinacharya@gmail.com"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? {'username': 'sachinacharya', 'domain': 'gmail'}

    pattern = "/path/name"
    string = "/path/name"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? {}

    pattern = "/"
    string = "/path/name"
    output = extract_route_pattern(pattern, string)
    print(output)  # ? None



from NetJin import WebServer, Request, Response

app = WebServer(True)


@app.route("/", methods=["GET"])
def index(_: Request, response: Response):
    response.render("index.html", name="Sachin Acharya")


@app.route("/about", methods=["GET"])
def about(_: Request, response: Response):
    response.render("about.html")


@app.route("/contact", methods=["GET"])
def contact(_: Request, response: Response):
    response.render("contact.html")


@app.route("/test/<name>/<userid>")
def test_name(request: Request, response: Response):
    print(request.user_parameters)
    response.send("Received")


@app.route("/users", methods=["GET", "POST"])
def get_users(_: Request, response: Response):
    response.send(
        [
            [
                {"name": "Sachin Acharya", "age": 24},
                {"name": "Nancy Jewel Mc.Donie", "age": 24},
                {"name": "Peyton List", "age": 23},
            ]
        ],
        None,
        {
            "Authorization": "Dickie",
            "Content-Type": "text/plain"
        }
    )


if __name__ == "__main__":
    app.run()